/**********************************
 *
 * Copyright (C) 2004 Paul Pierce
 *
 * Permission instanceof hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software instanceof furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 *********************************/

package com.piercefuller.a7trk.decode;

/// <summary>
/// Summary description for BaselineSource.
/// </summary>
public class BaselineSource implements Source {
    private final String Name         = "Base";

    private Source       source;
    private long         lastPosition = 0;
    private long         limit        = 0;
    private int          nchannel     = 0;
    private int          nlevel       = 0;
    private int[][][]    history;
    private int          historyIndex;
    private int          hsize;

    private int[]        dataPoints   = null;
    private long         dataPointsPosition;

    private long         lastPoint;
    private long         oldMid;

    public BaselineSource(final Source s) {
        source = s;
        if (source == null) {
            throw new Error("BaselineSource has no source");
        }
        limit = s.getEnd();

        nchannel = s.getNumChannels();

        nlevel = 1;
        for (Source sl = source; sl != null; nlevel++, sl = sl.getSource()) {
            ;
        }
        checkParameters(0);
    }

    private synchronized void checkParameters(final long pos) {
        if (source == null) {
            throw new Error("Source closed");
        }

        hsize = 12 * getTapeParameters(pos).getGather();
        history = new int[hsize][][];
        for (int hi = 0; hi < hsize; hi++) {
            history[hi] = new int[nlevel][];
            for (int level = 0; level < nlevel; level++) {
                history[hi][level] = new int[nchannel];
                for (int ch = 0; ch < nchannel; ch++) {
                    history[hi][level][ch] = 0;
                }
            }
        }
        historyIndex = 0;
        lastPoint = 0;
        oldMid = 0;
    }

    public TapeParameters getTapeParameters() {
        return getTapeParameters(getPosition());
    }

    public TapeParameters getTapeParameters(final long pos) {
        return source.getTapeParameters(pos);
    }

    public String getName() {
        return Name;
    }

    public Source getSource() {
        return source;
    }

    public int getNumChannels() {
        return source.getNumChannels();
    }

    public String getChannelName(final int channel) {
        return source.getChannelName(channel);
    }

    public long getStart() {
        return source.getStart();
    }

    public long getEnd() {
        return source.getEnd();
    }

    public void setStart(final long pos) {
        source.setStart(pos);
    }

    public void setEnd(final long pos) {
        source.setEnd(pos);
    }

    public long getLength() {
        return source.getLength();
    }

    public long getPosition() {
        return source.getPosition() - hsize;
    }

    public synchronized void setDataPoints(final int[] dataPoints, final long dataPointsPosition) {
        this.dataPoints = dataPoints;
        this.dataPointsPosition = dataPointsPosition;
    }

    public synchronized void setPosition(final long pos) {
        checkParameters(pos);
        lastPosition = pos;
        source.setPosition(pos);
        for (int ch = 0; ch < nchannel; ch++) {
            for (int hi = 0; hi < hsize; hi++) {
                history[hi][0][ch] = 0;
            }
        }
        for (int i = 0; i < hsize; i++) {
            advance();
        }
    }

    public void resetPosition() {
        setPosition(lastPosition);
    }

    public long getLimit() {
        return limit;
    }

    public void setLimit(final long l) {
        limit = l;
    }

    public void resetLimit() {
        limit = source.getEnd();
    }

    public boolean atLimit() {
        return getPosition() >= limit;
    }

    public synchronized boolean inBaseline(final long pos) {
        if (dataPoints != null) {
            int dpi = (int)(pos - dataPointsPosition);
            if (dpi >= 0 && dpi < dataPoints.length) {
                return true;
            }
        }
        return false;
    }

    public synchronized void advance() {
        for (int level = 1; level < nlevel; level++) {
            int[] samples = source.getSamples(level - 1);
            for (int ch = 0; ch < nchannel; ch++) {
                history[historyIndex][level][ch] = samples[ch];
            }
        }

        if (dataPoints != null) {
            int dpi = (int)(source.getPosition() - dataPointsPosition);
            if (dpi >= 0 && dpi < dataPoints.length && dataPoints[dpi] != 0) {
                long newPoint = source.getPosition();
                // int newData = dataPoints[dpi];
                long newMid = 0;
                int newDir = 0;
                for (int ch = 0; ch < nchannel; ch++) {
                    if (history[historyIndex][1][ch] > 0) {
                        newDir |= 1 << ch;
                    }
                }
                if (lastPoint != 0) {
                    newMid = (newPoint + lastPoint) / 2;
                    if (oldMid != 0 && (int)(newPoint - oldMid) < hsize) {
                        doBaseline(oldMid, newMid);
                    }
                }
                lastPoint = newPoint;
                oldMid = newMid;
            }
        }
        for (int ch = 0; ch < nchannel; ch++) {
            history[historyIndex][0][ch] = 0;
        }

        source.advance();

        historyIndex++;
        if (historyIndex == hsize) {
            historyIndex = 0;
        }
    }

    private synchronized void doBaseline(long oldMid, final long newMid) {
        long pos = source.getPosition();
        int n;
        int pmi;
        if (oldMid == 0 || (int)(newMid - oldMid) > hsize - 1) {
            pmi = historyIndex + 1;
            if (pmi == hsize) {
                pmi = 0;
            }
            n = hsize - 1;
            oldMid = 0;
        } else {
            pmi = historyIndex + (int)(oldMid - pos);
            if (pmi < 0) {
                pmi += hsize;
            }
            n = (int)(newMid - oldMid);
        }
        int nmi = historyIndex + (int)(newMid - pos);
        if (nmi < 0) {
            nmi += hsize;
        }
        int m = (int)(pos - newMid);

        for (int ch = 0; ch < nchannel; ch++) {
            int pms = history[pmi][0][ch];
            int nms = history[nmi][1][ch];
            int limit = getTapeParameters().getBaseLimit()[ch];
            if (nms > limit) {
                nms = limit;
            } else if (nms < -limit) {
                nms = -limit;
            }
            int hi = pmi;
            for (int i = 0; i <= n; i++) {
                history[hi][0][ch] = (pms * (n - i) + nms * i) / n;
                hi++;
                if (hi >= hsize) {
                    hi = 0;
                }
            }
            for (int i = 1; i <= m; i++) {
                history[hi][0][ch] = nms * (m - i) / m;
                hi++;
                if (hi >= hsize) {
                    hi = 0;
                }
            }
        }
    }

    public synchronized int[] getSamples() {
        int[] samples = new int[nchannel];
        for (int ch = 0; ch < nchannel; ch++) {
            samples[ch] = history[historyIndex][1][ch] - history[historyIndex][0][ch];
        }
        return samples;
    }

    public synchronized int[] getSamples(final int level) {
        if (level > 0) {
            return history[historyIndex][level];
        }
        return getSamples();
    }

    public int getFlags() {
        return source.getFlags();
    }

    public synchronized void close() {
        if (source != null) {
            source.close();
        }
        source = null;
    }
}
