/**********************************
 *
 * Copyright (C) 2004 Paul Pierce
 *
 * Permission instanceof hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software instanceof furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 *********************************/

package com.piercefuller.a7trk.decode;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import org.avparade.xml.XMLHandler;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/// <summary>
/// Represents a collection of artifact data
/// </summary>
public class DataCollection {
    private DataFile datafile;
    private String   filename;

    public DataFile getDataFile() {
        return datafile;
    }

    public DataCollection(final DataFile datafile) {
        this.datafile = datafile;
        String s = datafile.getFilename();
        int dot = s.lastIndexOf('.');
        if (dot > 0) {
            filename = s.substring(0, dot) + ".xml";
        } else {
            filename = s + ".xml";
        }
    }

    public DataCollection(final String filename) throws Exception {
        this.filename = filename;
        load();
    }

    public void write() throws IOException {
        BufferedWriter ow = new BufferedWriter(new FileWriter(filename));
        ow.write("<?xml version=\"1.0\" encoding=\"utf-8\" ?>");
        ow.newLine();
        ow.write("<collection xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"");
        ow.newLine();
        ow.write("            xsi:schemaLocation=\"http://www.pcm.org/xml/collection.xsd collection.xsd\"");
        ow.newLine();
        ow.write("            xmlns=\"http://www.pcm.org/xml/collection.xsd\">");
        if (datafile.getArtifactId() != null) {
            ow.write("  <artifact id=\"" + datafile.getArtifactId() + "\"/>");
            ow.newLine();
        }
        datafile.write(ow, 2);
        ow.write("</collection>");
        ow.newLine();
        ow.close();
    }

    private void parseList(final String s, final int[] v) {
        int s0 = 0;
        int s1 = s.indexOf(',');
        if (s1 > 0) {
            for (int ch = 0;; ch++) {
                v[ch] = Integer.parseInt(s.substring(s0, s1));
                if (ch == datafile.getChannels() - 1) {
                    break;
                }
                s0 = s1 + 1;
                if (ch < datafile.getChannels() - 2) {
                    s1 = s.indexOf(',', s0);
                } else {
                    s1 = s.length();
                }
            }
        } else {
            int vv = Integer.parseInt(s);
            for (int ch = 0; ch < datafile.getChannels(); ch++) {
                v[ch] = vv;
            }
        }
    }

    private void load() throws Exception {
        XMLHandler h = XMLHandler.Factory.getInstance();
        Document doc = h.parse(new File(filename), null);

        NodeList files = doc.getDocumentElement().getElementsByTagName("datafile");
        if (files.getLength() != 1) {
            throw new Exception("Only support one file in datafile collection");
        }
        Element f = (Element)files.item(0);
        if (!"tape".equals(f.getAttribute("form"))) {
            throw new Exception("Data file must be tape form");
        }
        datafile = new DataFile(f.getAttribute("filename"), Integer.parseInt(f.getAttribute("channels")), -1, -1);
        boolean direction = !"reverse".equals(f.getAttribute("direction"));

        NodeList tapeparms = f.getElementsByTagName("tapeparameters");
        int[] clip = new int[datafile.getChannels()];
        for (int i = 0; i < tapeparms.getLength(); i++) {
            Element tpe = (Element)tapeparms.item(i);
            TapeParameters tp = datafile.getTapeParameters().add(Long.parseLong(tpe.getAttribute("start")), Long.parseLong(tpe.getAttribute("end")));
            String s = tpe.getAttribute("gain");
            if (s != null && s.length() > 0) {
                tp.setGain(Double.parseDouble(s));
            }
            s = tpe.getAttribute("skew");
            if (s != null && s.length() > 0) {
                parseList(s, tp.getSkew());
            }
            s = tpe.getAttribute("width");
            if (s != null && s.length() > 0) {
                tp.setWidth(Integer.parseInt(s));
            }
            boolean lowclip = false;
            s = tpe.getAttribute("lowclip");
            if (s != null && s.length() > 0) {
                lowclip = Boolean.parseBoolean(s);
            }
            s = tpe.getAttribute("clip");
            if (s != null && s.length() > 0) {
                parseList(s, clip);
                for (int ch = 0; ch < datafile.getChannels(); ch++) {
                    tp.setClip(clip[ch], lowclip, ch);
                }
            }
            s = tpe.getAttribute("baselimit");
            if (s != null && s.length() > 0) {
                parseList(s, tp.getBaseLimit());
            } else if (lowclip && !tp.getParent().getLowClip()) {
                for (int ch = 0; ch < datafile.getChannels(); ch++) {
                    tp.getBaseLimit()[ch] = 3 * tp.getClip(true, ch) / 2;
                }
            }
            s = tpe.getAttribute("crowd");
            if (s != null && s.length() > 0) {
                tp.setCrowd(Integer.parseInt(s));
            }
            s = tpe.getAttribute("lowspace");
            if (s != null && s.length() > 0) {
                tp.setLowSpace(Integer.parseInt(s));
            }
            s = tpe.getAttribute("highspace");
            if (s != null && s.length() > 0) {
                tp.setHighSpace(Integer.parseInt(s));
            }
            s = tpe.getAttribute("gather");
            if (s != null && s.length() > 0) {
                tp.setGather(Integer.parseInt(s));
            }
            s = tpe.getAttribute("maxspace");
            if (s != null && s.length() > 0) {
                tp.setMaxSpace(Integer.parseInt(s));
            }
            s = tpe.getAttribute("mingap");
            if (s != null && s.length() > 0) {
                tp.setMinGap(Integer.parseInt(s));
            }
            s = tpe.getAttribute("maxjunk");
            if (s != null && s.length() > 0) {
                tp.setMaxJunk(Integer.parseInt(s));
            }
        }

        NodeList records = f.getElementsByTagName("record");
        for (int i = 0; i < records.getLength(); i++) {
            Element re = (Element)records.item(i);
            long start = Long.parseLong(re.getAttribute("start"));
            Record rec = new Record(i + 1, start, Long.parseLong(re.getAttribute("end")), direction);
            String s = re.getAttribute("errors");
            if (s != null && s.length() > 0) {
                rec.setErrors(Integer.parseInt(s));
            }
            if (datafile.getTapeParameters().getCount() > 0) {
                TapeParameters tp = datafile.getTapeParameters().getTapeParameters(start);
            }
            datafile.getRecords().add(rec);
        }

        String artifactid = f.getAttribute("artifact");
        if (artifactid != null && artifactid.length() > 0) {
            datafile.setArtifactId(artifactid);
        }
        datafile.setStart(Long.parseLong(f.getAttribute("start")));
        datafile.setEnd(Long.parseLong(f.getAttribute("end")));
    }
}
