/**********************************
 *
 * Copyright (C) 2004 Paul Pierce
 *
 * Permission instanceof hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software instanceof furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 *********************************/


package com.piercefuller.a7trk.decode;

    /// <summary>
    /// Summary description for LumpSource.
    /// </summary>
    public class LumpSource implements Source
    {
        private final String Name = "Lump";

        private Source source;
        private long lastPosition = 0;
        private long limit = 0;
        private int nchannel;
        private int[][] history;
        private int historyIndex = 0;
        private int[] sum;

        public LumpSource(Source source)
        {
            this.source = source;
            if (source == null) 
            {
                throw new Error("No source");
            }
            limit = source.getEnd();

            nchannel = source.getNumChannels();
            sum = new int[nchannel];
            for (int j = 0; j < nchannel; j++) 
            {
                sum[j] = 0;
            }
            setPosition(0);
        }

        public TapeParameters getTapeParameters() { return source.getTapeParameters(); }
        public TapeParameters getTapeParameters(long pos) { return source.getTapeParameters(pos); }

        private void checkParameter(long pos) 
        {
            if (pos < 0) 
            {
                pos = 0;
            }
            TapeParameters tp = source.getTapeParameters(pos);
            if (history == null || history.length != tp.getWidth())
            {
                history = new int[tp.getWidth()][];
                for (int i = 0; i < history.length; i++) 
                {
                    history[i] = new int[nchannel];
                }
            }
        }

        public String getName() { return Name; }

        public Source getSource() { return source; }

        public int getNumChannels() { return source.getNumChannels(); }

        public String getChannelName(int channel) { return source.getChannelName(channel); }

        public long getStart() { return source.getStart(); }

        public long getEnd() { return source.getEnd(); }

        public void setStart(long pos) { source.setStart(pos); }

        public void setEnd(long pos) { source.setEnd(pos); }

        public long getLength() 
        {
            return source.getLength();
        }

        public long getPosition()
        {
            return source.getPosition() - history.length/2;
        }

        public void setPosition(long pos)
        {
            if (source == null) 
            {
                throw new Error("Source is closed");
            }
            checkParameter(pos);
            
            source.setPosition(pos - history.length/2 - 1);

            for (int i = 0; i < nchannel; i++) 
            {
                sum[i] = 0;
            }
            for (historyIndex = 0; historyIndex < history.length; historyIndex++)
            {
                for (int ch = 0; ch < nchannel; ch++) 
                {
                    history[historyIndex][ch] = 0;
                }
            }
            historyIndex = 0;
            for (; getPosition() < pos; ) 
            {
                advance();
            }
            lastPosition = getPosition();
        }

        public void resetPosition()
        {
            setPosition(lastPosition);
        }

        public long getLimit()
        {
            return limit;
        }

        public void setLimit(long l)
        {
            limit = l;
        }

        public void resetLimit()
        {
            limit = source.getEnd();
        }

        public boolean atLimit()
        {
            return getPosition() >= limit;
        }

        public void advance()
        {
            historyIndex++;
            if (historyIndex == history.length) 
            {
                historyIndex = 0;
            }
            int[] samples = source.getSamples();
            for (int i = 0; i < nchannel; i++) 
            {
                sum[i] -= history[historyIndex][i];
                history[historyIndex][i] = samples[i];
                sum[i] += history[historyIndex][i];
            }
            source.advance();
        }

        public int[] getSamples()
        {
            return sum;
        }

        public int[] getSamples(int level)
        {
            if (level > 1) 
            {
                throw new Error("GetSamples level > 1 in LumpSource");
            }
            if (level == 0) 
            {
                return getSamples();
            }

            int middle = historyIndex - history.length/2;
            if (middle < 0) 
            {
                middle += history.length;
            }
            if (history[middle] != null) 
            {
                return history[middle];
            }

            int[] samples = new int[nchannel];
            for (int i = 0; i < nchannel; i++) 
            {
                samples[i] = 0;
            }
            return samples;
        }

        public int getFlags()
        {
            return source.getFlags();
        }

        public void close()
        {
            if (source != null) 
            {
                source.close();
            }
            source = null;
        }
    }
