/**********************************
 *
 * Copyright (C) 2004, 2013 Paul Pierce
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 *********************************/

package com.piercefuller.a7trk.decode;

import java.io.IOException;
import java.util.Arrays;

public abstract class RawSource implements Source {
    public static final int     FlagBOT      = 0x0100;
    public static final int     FlagEOT      = 0x0200;
    public static final int     FlagForward  = 0x0001;
    public static final int     FlagReverse  = 0x0002;
    private static final String Name         = "Raw";
    private static final int    MaxSkew      = 10;

    private DataFile            datafile;
    private long                startPos     = 0;
    private long                endPos       = 0;
    private long                position     = 0;
    private long                lastPosition = 0;
    private long                limit        = 0;
    private int[][]             history      = new int[MaxSkew][];
    private int                 historyIndex = 0;

    public RawSource(final DataFile datafile) {
        this.datafile = datafile;
    }

    public void initialize() throws Exception {
        limit = getLength();
        if (datafile.getEnd() > 0) {
            startPos = datafile.getStart();
            endPos = datafile.getEnd();
        } else {
            startPos = 0;
            endPos = getLength();
            datafile.setStart(startPos);
            datafile.setEnd(endPos);
        }

        for (int i = 0; i < MaxSkew; i++) {
            history[i] = new int[getNumChannels()];
            for (int j = 0; j < getNumChannels(); j++) {
                history[i][j] = 0;
            }
        }

        internalSetPosition(0);
    }

    @Override
    public abstract Object clone();

    public String getName() {
        return Name;
    }

    public Source getSource() {
        return null;
    }

    public TapeParameters getTapeParameters() {
        return getTapeParameters(getPosition());
    }

    public TapeParameters getTapeParameters(final long pos) {
        return datafile.getTapeParameters().getTapeParameters(pos);
    }

    public long getStart() {
        return startPos;
    }

    public long getEnd() {
        return endPos;
    }

    public synchronized void setStart(final long pos) {
        try {
            datafile.setStart(pos);
            startPos = pos;
        } catch (Exception e) {
            System.err.println(e);
        }
    }

    public synchronized void setEnd(final long pos) {
        try {
            datafile.setEnd(pos);
            endPos = pos;
        } catch (Exception e) {
            System.err.println(e);
        }
    }

    public long getPosition() {
        return position;
    }

    public synchronized void setPosition(final long pos) {
        try {
            lastPosition = pos;
            internalSetPosition(pos);
        } catch (Exception e) {
            System.err.println(e);
        }
    }

    public synchronized void resetPosition() {
        try {
            internalSetPosition(lastPosition);
        } catch (Exception e) {
            System.err.println(e);
        }
    }

    public long getLimit() {
        return limit;
    }

    public void setLimit(final long l) {
        limit = l;
    }

    public void resetLimit() {
        limit = getLength();
    }

    public boolean atLimit() {
        return position >= limit;
    }

    protected void internalSetPosition(final long pos) throws Exception {
        for (int i = 0; i < history.length; i++) {
            for (int j = 0; j < getNumChannels(); j++) {
                history[i][j] = 0;
            }
        }

        position = pos - MaxSkew;
        long n = position / 2;
        if (n < 0) {
            n = 0;
        }
        sampleSeek(n);
        for (historyIndex = 0; position < pos; historyIndex++) {
            if (position >= 0 && position >= lastPosition && position < limit) {
                readFlat(history[historyIndex]);
            } else {
                Arrays.fill(history[historyIndex], 0);
            }
            position++;
        }
        historyIndex--;
    }

    public abstract void sampleSeek(long n) throws IOException;

    public abstract void readFlat(final int[] samples) throws Exception;

    /**
     * @return the datafile
     */
    public DataFile getDatafile() {
        return datafile;
    }

    /**
     * @param datafile the datafile to set
     */
    public void setDatafile(final DataFile datafile) {
        this.datafile = datafile;
    }

    public synchronized void advance() {
        historyIndex++;
        if (historyIndex == MaxSkew) {
            historyIndex = 0;
        }
        try {
            if (position >= 0 && position >= lastPosition && position < limit) {
                readFlat(history[historyIndex]);
            } else {
                Arrays.fill(history[historyIndex], 0);
            }
            position++;
        } catch (Exception e) {
            System.err.println(e);
        }
    }

    public synchronized int[] getSamples() {
        int[] samples = new int[getNumChannels()];

        for (int i = 0; i < getNumChannels(); i++) {
            int skewIndex = historyIndex;
            TapeParameters tp = getTapeParameters();
            if (tp != null) {
                skewIndex -= tp.getSkew()[i];
            }
            if (skewIndex < 0) {
                skewIndex += MaxSkew;
            }
            samples[i] = history[skewIndex][i];
        }
        return samples;
    }

    public synchronized int[] getSamples(final int level) {
        if (level > 0) {
            throw new Error("GetSamples level > 0 in RawSource");
        }
        return getSamples();
    }

    public static RawSource getInstance(final DataFile datafile) throws Exception {
        String filename = datafile.getFilename();
        if (filename.toLowerCase().endsWith(".dat")) {
            return new DatRawSource(datafile);
        } else if (filename.toLowerCase().endsWith(".hpf")) {
            return new HpfRawSource(datafile);
        } else {
            throw new Exception("Unsupported file extension on file \"" + filename + "\"");
        }
    }
}
