/**********************************
 *
 * Copyright (C) 2004 Paul Pierce
 *
 * Permission instanceof hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software instanceof furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 *********************************/

package com.piercefuller.a7trk.decode;

import java.io.BufferedWriter;
import java.io.IOException;

/// <summary>
/// Summary description for TapeParameters.
/// </summary>
public class TapeParameters {
    private TapeParametersList tpl;
    private TapeParameters     parent    = null;
    private double             gain      = 1200;
    private long               start     = 0;
    private long               end       = 0;
    private int                nchannel  = 0;
    private int[]              skew      = null;
    private int                width     = 9;
    private boolean            lowClip   = false;
    private int[]              clip      = null;
    private int[]              limit     = null;
    private int                lowspace  = 35;
    private int                highspace = 150;
    private int                crowd     = 13;
    private int                gather    = 3;
    private int                maxspace  = 300;
    private int                mingap    = 6000;
    private int                maxjunk   = 5;

    private int getDefaultClip() {
        return width * 2048 / 8;
    }

    public TapeParameters(final int nchannel, final long start, final long end, final TapeParameters parent, final TapeParametersList tpl) {
        this.start = start;
        this.end = end;
        this.parent = parent;
        this.nchannel = nchannel;
        this.tpl = tpl;
        skew = new int[nchannel];
        clip = new int[nchannel];
        limit = new int[nchannel];

        for (int i = 0; i < nchannel; i++) {
            skew[i] = 0;
            clip[i] = getDefaultClip();
            limit[i] = clip[i];
        }
        if (parent != null) {
            gain = parent.gain;
            width = parent.width;
            lowClip = parent.lowClip;
            for (int ch = 0; ch < nchannel; ch++) {
                clip[ch] = parent.clip[ch];
                skew[ch] = parent.skew[ch];
                limit[ch] = parent.limit[ch];
            }
            lowspace = parent.lowspace;
            highspace = parent.highspace;
            crowd = parent.crowd;
            gather = parent.gather;
            maxspace = parent.maxspace;
            mingap = parent.mingap;
            maxjunk = parent.maxjunk;
        }
    }

    public long getStart() {
        return start;
    }

    public void setStart(final long value) throws Exception {
        if (value <= end && (parent == null || value >= parent.start && value <= parent.end)) {
            start = value;
            tpl.invalidate();
        } else {
            throw new Exception("Invalid start position");
        }
    }

    public long getEnd() {
        return end;
    }

    public void setEnd(final long value) throws Exception {
        if (value >= start && (parent == null || value >= parent.start && value <= parent.end)) {
            end = value;
            tpl.invalidate();
        } else {
            throw new Exception("Invalid end position");
        }
    }

    public boolean getLowClip() {
        return lowClip;
    }

    public void setLowClip(final boolean value) {
        if (canChangeLowClip()) {
            lowClip = value;
        }
    }

    public boolean canChangeLowClip() {
        if (parent == null) {
            return false;
        }
        return !parent.lowClip;
    }

    public TapeParameters getParent() {
        return parent;
    }

    public void setParent(final TapeParameters value) {
        parent = value;
    }

    /**
     * @return the gain
     */
    public double getGain() {
        return gain;
    }

    /**
     * @param gain the gain to set
     */
    public void setGain(final double gain) {
        this.gain = gain;
    }

    public int getWidth() {
        return width;
    }

    public void setWidth(final int value) {
        width = value + ((value + 1) % 2);
    }

    public int getLowSpace() {
        return lowspace;
    }

    public void setLowSpace(final int value) {
        lowspace = value;
    }

    public int getHighSpace() {
        return highspace;
    }

    public void setHighSpace(final int value) {
        highspace = value;
    }

    public int getCrowd() {
        return crowd;
    }

    public void setCrowd(final int value) {
        crowd = value;
    }

    public int getGather() {
        return gather;
    }

    public void setGather(final int value) {
        gather = value;
    }

    public int getMaxSpace() {
        return maxspace;
    }

    public void setMaxSpace(final int value) {
        maxspace = value;
    }

    public int getMinGap() {
        return mingap;
    }

    public void setMinGap(final int value) {
        mingap = value;
    }

    public int getMaxJunk() {
        return maxjunk;
    }

    public void setMaxJunk(final int value) {
        maxjunk = value;
    }

    public int[] getSkew() {
        return skew;
    }

    public int[] getBaseLimit() {
        return limit;
    }

    public int getClip(final boolean low, final int channel) {
        if (!lowClip || low) {
            return clip[channel];
        }
        for (TapeParameters tp = parent; tp != null; tp = tp.parent) {
            if (!tp.lowClip) {
                return tp.clip[channel];
            }
        }
        return getDefaultClip();
    }

    public void setClip(final int i, final int channel) {
        clip[channel] = i;
    }

    public void setClip(final int i, final boolean low, final int channel) {
        clip[channel] = i;
        lowClip = low;
    }

    public void write(final BufferedWriter ow, final int indent) throws IOException {
        String inds = "                    ".substring(0, indent);
        String lasts = inds;
        ow.write(inds + "<tapeparameters start=\"" + start + "\" end=\"" + end + "\"");
        ow.newLine();

        if (parent == null || parent.gain != gain) {
            String s = " gain=\"" + gain + "\"";
            if (lasts.length() + s.length() > 72) {
                ow.write(lasts);
                ow.newLine();
                lasts = inds;
            }
            lasts += s;
        }

        boolean parentDiff = true;
        boolean selfSame = true;
        for (int i = 1; i < nchannel; i++) {
            if (skew[i] != skew[0]) {
                selfSame = false;
            }
        }
        if (parent != null) {
            parentDiff = false;
            for (int i = 0; i < nchannel; i++) {
                if (skew[i] != parent.skew[i]) {
                    parentDiff = true;
                    break;
                }
            }
        }
        if (parentDiff) {
            String s = " skew=\"";
            for (int i = 0; i < nchannel; i++) {
                s += skew[i];
                if (selfSame) {
                    break;
                }
                if (i < nchannel - 1) {
                    s += ",";
                }
            }
            s += "\"";
            if (lasts.length() + s.length() > 72) {
                ow.write(lasts);
                ow.newLine();
                lasts = inds;
            }
            lasts += s;
        }

        if (parent == null || parent.width != width) {
            String s = " width=\"" + width + "\"";
            if (lasts.length() + s.length() > 72) {
                ow.write(lasts);
                ow.newLine();
                lasts = inds;
            }
            lasts += s;
        }

        parentDiff = true;
        selfSame = true;
        for (int i = 1; i < nchannel; i++) {
            if (clip[i] != clip[0]) {
                selfSame = false;
            }
        }
        if (parent != null && parent.lowClip == lowClip) {
            parentDiff = false;
            for (int i = 0; i < nchannel; i++) {
                if (clip[i] != parent.clip[i]) {
                    parentDiff = true;
                    break;
                }
            }
        }
        if (parentDiff) {
            String s = " lowclip=\"" + lowClip + "\" clip=\"";
            for (int i = 0; i < nchannel; i++) {
                s += clip[i];
                if (selfSame) {
                    break;
                }
                if (i < nchannel - 1) {
                    s += ",";
                }
            }
            s += "\"";
            if (lasts.length() + s.length() > 72) {
                ow.write(lasts);
                ow.newLine();
                lasts = inds;
            }
            lasts += s;
        }

        parentDiff = true;
        selfSame = true;
        for (int i = 1; i < nchannel; i++) {
            if (limit[i] != limit[0]) {
                selfSame = false;
            }
        }
        if (parent != null) {
            parentDiff = false;
            for (int i = 0; i < nchannel; i++) {
                if (limit[i] != parent.limit[i]) {
                    parentDiff = true;
                    break;
                }
            }
        }
        if (parentDiff) {
            String s = " baselimit=\"";
            for (int i = 0; i < nchannel; i++) {
                s += limit[i];
                if (selfSame) {
                    break;
                }
                if (i < nchannel - 1) {
                    s += ",";
                }
            }
            s += "\"";
            if (lasts.length() + s.length() > 72) {
                ow.write(lasts);
                ow.newLine();
                lasts = inds;
            }
            lasts += s;
        }

        if (parent == null || parent.crowd != crowd) {
            String s = " crowd=\"" + crowd + "\"";
            if (lasts.length() + s.length() > 72) {
                ow.write(lasts);
                ow.newLine();
                lasts = inds;
            }
            lasts += s;
        }

        if (parent == null || parent.lowspace != lowspace) {
            String s = " lowspace=\"" + lowspace + "\"";
            if (lasts.length() + s.length() > 72) {
                ow.write(lasts);
                ow.newLine();
                lasts = inds;
            }
            lasts += s;
        }

        if (parent == null || parent.highspace != highspace) {
            String s = " highspace=\"" + highspace + "\"";
            if (lasts.length() + s.length() > 72) {
                ow.write(lasts);
                ow.newLine();
                lasts = inds;
            }
            lasts += s;
        }

        if (parent == null || parent.gather != gather) {
            String s = " gather=\"" + gather + "\"";
            if (lasts.length() + s.length() > 72) {
                ow.write(lasts);
                ow.newLine();
                lasts = inds;
            }
            lasts += s;
        }

        if (parent == null || parent.maxspace != maxspace) {
            String s = " maxspace=\"" + maxspace + "\"";
            if (lasts.length() + s.length() > 72) {
                ow.write(lasts);
                ow.newLine();
                lasts = inds;
            }
            lasts += s;
        }

        if (parent == null || parent.mingap != mingap) {
            String s = " mingap=\"" + mingap + "\"";
            if (lasts.length() + s.length() > 72) {
                ow.write(lasts);
                ow.newLine();
                lasts = inds;
            }
            lasts += s;
        }

        if (parent == null || parent.maxjunk != maxjunk) {
            String s = " maxjunk=\"" + maxjunk + "\"";
            if (lasts.length() + s.length() > 72) {
                ow.write(lasts);
                ow.newLine();
                lasts = inds;
            }
            lasts += s;
        }

        ow.write(lasts + " />");
        ow.newLine();
    }
}
