/**
 * IsoTime.java
 *
 * Copyright (C) 2007 Paul R. Pierce
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see http://www.gnu.org/licenses/gpl.txt
 */
package org.avparade.xml;

import java.util.*;

/**
 * org.avparade.xml.IsoTime
 *
 */
public class IsoTime implements Comparable<IsoTime> {
	private Calendar calendar;
	private IsoDuration duration = null;

	public IsoTime() {
		this(System.currentTimeMillis());
	}
	
	/**
	 * @param calendar
	 */
	public IsoTime(Calendar calendar) {
		this.calendar = calendar;
	}
	
	public IsoTime(long millis) {
		calendar = Calendar.getInstance();
		calendar.setTimeZone(TimeZone.getTimeZone("GMT"));
		calendar.setTimeInMillis(millis);
	}
	
	public IsoTime(IsoTime start, IsoTime end) {
		calendar = start.getCalendar();
		duration = new IsoDuration(start, end);
	}
	
	private class StringPosition {
		String string;
		int index;

		public StringPosition(String string, int index) {
			this.string = string;
			this.index = index;
		}
		public char thisChar() {
			return Character.toUpperCase(string.charAt(index));
		}
		public char nextChar() {
			return Character.toUpperCase(string.charAt(index + 1));
		}
		public boolean atBeginning() {
			return index == 0;
		}
		public boolean atEnd() {
			return index == string.length();
		}
		public boolean has(int digits) {
			return index + digits <= string.length();
		}
		public String take(int digits) {
			index += digits;
			return string.substring(index - digits, index);
		}
		public void skip(int digits) {
			index += digits;
		}
		public String remaining() {
			return string.substring(index);
		}
	}

	/**
	 * @param calendar
	 */
	public IsoTime(String iso) {
		if (iso.length() == 20 &&
				iso.charAt(4) == '-' && iso.charAt(7) == '-' &&
				iso.charAt(10) == 'T' &&
				iso.charAt(13) == ':' && iso.charAt(16) == ':' &&
				iso.charAt(19) == 'Z' ) {
			calendar = Calendar.getInstance(TimeZone.getTimeZone("UTC"));
			calendar.set(Integer.parseInt(iso.substring(0, 4)),
					Integer.parseInt(iso.substring(5, 7)) + (Calendar.JANUARY - 1),
					Integer.parseInt(iso.substring(8, 10)),
					Integer.parseInt(iso.substring(11, 13)),
					Integer.parseInt(iso.substring(14, 16)),
					Integer.parseInt(iso.substring(17, 19)));
		}
		int division = iso.indexOf('/');
		if (division < 0) {
			division = -1;
			for (;;) {
				division = iso.indexOf('-', division + 1);
				if (division < 0) {
					break;
				}
				if (division + 1 < iso.length() && iso.charAt(division + 1) == '-') {
					break;
				}
			}
		}
		if (division > 0) {
			String firstPart = iso.substring(0, division);
			if (iso.charAt(division) == '-') {
				++division;
			}
			String secondPart = iso.substring(division + 1);
			if (Character.toUpperCase(firstPart.charAt(0)) == 'P') {
				duration = new IsoDuration(firstPart, false);
				init(secondPart);
			} else {
				init(firstPart);
				duration = new IsoDuration(secondPart, true);
			}
		} else {
			init(iso);
		}
	}

	private void init(String iso) {
		for (int i = 0; i < iso.length(); i++) {
			char c = Character.toUpperCase(iso.charAt(i));
			if (c == 'Z' && i == iso.length()-1) {
				--i;
			} else if (!Character.isDigit(c)) {
				break;
			} else if (i < iso.length() - 1) {
				continue;
			}
			if (c == 'Z' || i+1 == 4 || i+1 == 8) {
				calendar = Calendar.getInstance(TimeZone.getTimeZone("UTC"));
			} else {
				calendar = Calendar.getInstance();
			}
			calendar.set(Calendar.SECOND, 0);
			calendar.set(Calendar.MILLISECOND, 0);
			switch (i + 1) {
			case 2:
				calendar.set(Calendar.HOUR_OF_DAY, Integer.parseInt(iso.substring(0, 2)));
				calendar.set(Calendar.MINUTE, 0);
				return;
			case 4:
				if (c == 'Z') {
					break;
				}
				calendar.set(Calendar.YEAR, Integer.parseInt(iso));
				calendar.set(Calendar.MONTH, Calendar.JANUARY);
				calendar.set(Calendar.DAY_OF_MONTH, 1);
				calendar.set(Calendar.HOUR_OF_DAY, 0);
				calendar.set(Calendar.MINUTE, 0);
				return;
			case 6:
				calendar.set(Calendar.HOUR_OF_DAY, Integer.parseInt(iso.substring(0, 2)));
				calendar.set(Calendar.MINUTE, Integer.parseInt(iso.substring(2, 4)));
				calendar.set(Calendar.SECOND, Integer.parseInt(iso.substring(4, 6)));
				return;
			case 8:
				if (c == 'Z') {
					break;
				}
				calendar.set(Calendar.YEAR, Integer.parseInt(iso.substring(0, 4)));
				calendar.set(Calendar.MONTH, Integer.parseInt(iso.substring(4, 6) + (Calendar.JANUARY - 1)));
				calendar.set(Calendar.DAY_OF_MONTH, Integer.parseInt(iso.substring(6)));
				calendar.set(Calendar.HOUR_OF_DAY, 0);
				calendar.set(Calendar.MINUTE, 0);
				return;
			}
			break;
		}

		calendar = Calendar.getInstance(TimeZone.getTimeZone("UTC"));
		Exception ex = null;
		try {
			for (;;) { // This is not actually a loop, just something to break out of
				char c;
				StringPosition p = new StringPosition(iso, 0);
				
				calendar.set(Calendar.SECOND, 0);
				calendar.set(Calendar.MILLISECOND, 0);
				boolean going = getNumber(Calendar.YEAR, p, 4, 0, 0, true);
				if (getNumber(Calendar.MONTH, p, 2, '-', Calendar.JANUARY - 1, p.atBeginning())) {
					going = true;
				} else if (going) {
					calendar.set(Calendar.MONTH, Calendar.JANUARY);
				}
				if (getNumber(Calendar.DAY_OF_MONTH, p, 2, '-', 0, p.atBeginning())) {
					going = true;
				} else if (going) {
					calendar.set(Calendar.DAY_OF_MONTH, 1);
				}
				if (p.atEnd()) {
					if (going) {
						calendar.set(Calendar.HOUR_OF_DAY, 0);
						calendar.set(Calendar.MINUTE, 0);
					}
					return;
				}
				if (p.thisChar() != 'T') {
					break;
				}
				p.skip(1);

				if (Character.isDigit(p.thisChar())) {
					if (getNumber(Calendar.HOUR_OF_DAY, p, 2, 0, 0, false)) {
						going = true;
					} else if (going) {
						calendar.set(Calendar.HOUR_OF_DAY, 0);
					}
					if (getNumber(Calendar.MINUTE, p, 2, ':', 0, false)) {
						going = true;
					} else if (going) {
						calendar.set(Calendar.MINUTE, 0);
					}
					if (getNumber(Calendar.SECOND, p, 2, ':', 0, false)) {
						if (!p.atEnd() && p.thisChar() == '.') {
							p.skip(1);
							int j = p.index;
							for (; j < iso.length() && Character.isDigit(iso.charAt(j)); j++);
							j -= p.index;
							if (j == 0) {
								break;
							}
							int m = Integer.parseInt(p.take(j)) * 1000;
							for (; j > 0; j--) {
								m /= 10;
							}
							calendar.set(Calendar.MILLISECOND, m);
						}
					}
				}
				if (!p.atEnd()) {
					c = p.thisChar();
				} else {
					c = ' ';
				}
				if (c == 'Z') {
					p.skip(1);
				} else if (c == '+' || c == '-') {
					if (!p.has(5)) {
						break;
					}
					calendar.add(Calendar.MINUTE, Integer.parseInt(p.take(5)));
				} else {
					Calendar tc = Calendar.getInstance();
					tc.set(Calendar.MILLISECOND, calendar.get(Calendar.MILLISECOND));
					tc.set(Calendar.SECOND, calendar.get(Calendar.SECOND));
					tc.set(Calendar.MINUTE, calendar.get(Calendar.MINUTE));
					tc.set(Calendar.HOUR_OF_DAY, calendar.get(Calendar.HOUR_OF_DAY));
					tc.set(Calendar.DAY_OF_MONTH, calendar.get(Calendar.DAY_OF_MONTH));
					tc.set(Calendar.MONTH, calendar.get(Calendar.MONTH));
					tc.set(Calendar.YEAR, calendar.get(Calendar.YEAR));
					calendar.setTimeInMillis(tc.getTimeInMillis());
				}
				if (p.atEnd()) {
					return;
				}
				break;
			}
		} catch (Exception e) {
			ex = e;
		}
		NumberFormatException em = new NumberFormatException("Incompatible ISO time string <" + iso + ">");
		if (ex != null) {
			em.initCause(ex);
		}
		throw em;
	}

	private boolean getNumber(int field, StringPosition p, int digits, int separator, int adjust, boolean beforeT) {
		if (beforeT || separator != 0) {
			if (p.atEnd()) {
				return false;
			}
			char c = p.thisChar();
			if (beforeT && c == 'T') {
				return false;
			}
			if (separator != 0 && c == separator) {
				p.skip(1);
			}
		}
		if (!p.has(digits)) {
			throw new NumberFormatException("Too few digits");
		}
		calendar.set(field, Integer.parseInt(p.take(digits)) + adjust);
		return true;
	}
	
	/**
	 * @return the calendar
	 */
	public Calendar getCalendar() {
		return (Calendar)calendar.clone();
	}

	/**
	 * 
	 * @return the start time
	 */
	public Calendar getStart() {
		if (duration == null || duration.isForward()) {
			return getCalendar();
		}
		return duration.getEndTime(this).getCalendar();
	}

	/**
	 * 
	 * @return the end time
	 */
	public Calendar getEnd() {
		if (duration == null || !duration.isForward()) {
			return getCalendar();
		}
		return duration.getEndTime(this).getCalendar();
	}

	/**
	 * 
	 * @return the time in milliseconds
	 */
	public long getMillis() {
		return calendar.getTimeInMillis();
	}
	
	/**
	 * @return Returns the duration.
	 */
	public IsoDuration getDuration() {
		if (duration == null) {
			return new IsoDuration();
		}
		return (IsoDuration)duration.clone();
	}

	/* (non-Javadoc)
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	
	public int compareTo(IsoTime o) {
		return calendar.compareTo(o.calendar);
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	
	public boolean equals(Object obj) {
		if (obj instanceof IsoTime) {
			return compareTo((IsoTime)obj) == 0;
		}
		return false;
	}

	/**
	 * @return the time in ISO format with dashes and colons
	 */
	public String toString() {
		String dstring = "";
		if (duration != null) {
			dstring = "/" + duration.toString();
		}
		return Integer.toString(calendar.get(Calendar.YEAR)) + '-' +
			Integer.toString(calendar.get(Calendar.MONTH) - Calendar.JANUARY + 1 + 100).substring(1) + '-' +
			Integer.toString(calendar.get(Calendar.DAY_OF_MONTH) + 100).substring(1) + 'T' +
			Integer.toString(calendar.get(Calendar.HOUR_OF_DAY) + 100).substring(1) + ':' +
			Integer.toString(calendar.get(Calendar.MINUTE) + 100).substring(1) + ':' +
			Integer.toString(calendar.get(Calendar.SECOND) + 100).substring(1) + 'Z' + 
			dstring;
	}
}
