/**
 * XmlLog.java
 *
 * Copyright (C) 2007 Paul R. Pierce
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see http://www.gnu.org/licenses/gpl.txt
 */
package org.avparade.xml;

import java.util.*;
import java.io.*;
import org.w3c.dom.*;

/**
 * org.avparade.xml.mediagarden.XmlLog
 *
 */
public class XmlLog implements Log {
	public static final String LogNamespace = "http://www.purpos.org/xml/log.xsd";
	
	private File logFile;
	private Document doc;
	private XMLHandler xmlHandler;
	boolean firstFlush = true;
	boolean clean = false;

	/**
	 * @param logFile
	 * @param xmlHandler
	 */
	public XmlLog(File logFile, XMLHandler xmlHandler) throws Exception {
		this.logFile = logFile;
		this.xmlHandler = xmlHandler;
		if (logFile.exists()) {
			doc = xmlHandler.parse(logFile, null);
		} else {
			doc = xmlHandler.parse("<?xml version=\"1.1\" encoding=\"UTF-8\"?><log xmlns=\"" + LogNamespace + "\"/>", null);
		}
	}

	/* (non-Javadoc)
	 * @see org.avparade.xml.Log#add(java.lang.String)
	 */
	public void add(String entry) {
		Element note = doc.createElement(Log.NoteEntry);
		note.appendChild(doc.createTextNode(entry));
		add(note);
	}

	/* (non-Javadoc)
	 * @see org.avparade.xml.Log#add(org.w3c.dom.Element)
	 */
	public void add(Element entry) {
		if (entry.getOwnerDocument() != doc) {
			entry = (Element)doc.importNode(entry, true);
		}
		entry.setAttribute(Log.Time, new IsoTime().toString());
		doc.getDocumentElement().appendChild(entry);
		clean = false;
	}

	/* (non-Javadoc)
	 * @see org.avparade.xml.Log#mostRecent(java.lang.String, java.lang.String)
	 */
	public Element mostRecent(String entryName, String id) {
		for (Node node = doc.getDocumentElement().getLastChild(); node != null; node = node.getPreviousSibling()) {
			if (node instanceof Element) {
				Element elt = (Element)node;
				if (elt.getTagName().equals(entryName) && elt.getAttribute(Log.Id).equals(id)) {
					return elt;
				}
			}
		}
		return null;
	}

	/* (non-Javadoc)
	 * @see org.avparade.xml.Log#whatTime(org.w3c.dom.Element)
	 */
	public Calendar whatTime(Element entry) {
		String iso = entry.getAttribute(Log.Time);
		if (iso != null && iso.length() > 0) {
			try {
				Calendar cal = Calendar.getInstance();
				cal.set(Integer.parseInt(iso.substring(0, 4)),
						Integer.parseInt(iso.substring(5, 7)),
						Integer.parseInt(iso.substring(8, 10)),
						Integer.parseInt(iso.substring(11, 13)),
						Integer.parseInt(iso.substring(14, 16)),
						Integer.parseInt(iso.substring(17, 19)));
				return cal;
			} catch (NumberFormatException e) {
				System.err.println("XmlLog parsing time string " + iso + " - " + e);
			}
		}
		return null;
	}

	/* (non-Javadoc)
	 * @see org.purpos.avmedia.Log#getDocument()
	 */
	public Document getDocument() {
		return doc;
	}

	/* (non-Javadoc)
	 * @see org.avparade.xml.Log#flush()
	 */
	public void flush() throws Exception {
		if (clean) {
			return;
		}
		boolean ok = true;
		if (firstFlush) {
			if (logFile.exists()) {
				File older = new File(logFile.getCanonicalPath() + ".older");
				if (older.exists()) {
					File oldest = new File(logFile.getCanonicalPath() + ".oldest");
					if (!oldest.exists()) {
						ok &= older.renameTo(oldest);
					} else {
						ok &= older.delete();
					}
				}
				if (ok) {
					ok &= logFile.renameTo(older);
				}
			}
			firstFlush = false;
		} else if (logFile.exists()) {
			File old = new File(logFile.getCanonicalPath() + ".old");
			if (old.exists()) {
				ok &= old.delete();
			}
			if (ok) {
				ok &= logFile.renameTo(old);
			}
		}
		File file = logFile;
		if (!ok) {
			file = new File(logFile.getCanonicalPath() + ".new");
		}
		xmlHandler.write(file, doc, null);
		clean = true;
	}
}
